/***************************************************************************
 * athcool (Athlon Powersaving utility)

athcool is a small utility, enabling/disabling Powersaving mode
for AMD Athlon/Duron processors.

Since enabling Powersaving mode, you can save power consumption, 
lower CPU temprature when CPU is idle.

Powersaving works if your kernel support ACPI (APM not work),
because athcool only set/unset "Disconnect enable when STPGNT detected"
bits in the Northbridge of Chipset.
To really save power, someone has to send the STPGNT signal when idle.
This is done by the ACPI subsystem when C2 state enterd.

!!!WARNING!!!
Depending your motherboard and/or hardware components, 
enabling Athlon powersaving mode sometimes causes that

 * noisy sound playback
 * slowdown harddisk performance
 * system hangup or instable

If you met those problems, you should not use athcool.
Please use athcool at your own risk.

   Osamu Kayasono <jacobi@jcom.home.ne.jp>
   http://members.jcom.home.ne.jp/jacobi/linux/softwares.html

  (Related links)
 * CoolOn Project [ written in Japanese ]
	http://homepage2.nifty.com/coolon/
 * pciutils
	http://atrey.karlin.mff.cuni.cz/~mj/pciutils.shtml
 * Athlon Powersaving HOWTO
	http://cip.uni-trier.de/nofftz/linux/Athlon-Powersaving-HOWTO.html
 * LVCool
	http://vcool.occludo.net/

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <sys/io.h>
#include <string.h>
#include <pci/pci.h>
#include  "athcool.h"
#include  "chips.h"

#define ATHLONPM_ON	1
#define ATHLONPM_OFF	2
#define ATHLONPM_STAT	3
#define ATHLONPM_USAGE	4
#define ATHLONPM_LIST	5

extern struct pci_filter	filter;
extern struct pci_access	*pacc;
extern struct device	*scan_devices( void );

static void	print_title( void )
{
  fprintf( stderr, PACKAGE " version " VERSION "\n" );
  fprintf( stderr, " enabling/disabling Athlon Powersaving mode\n\n");
}

static void	list_chipsets(void)
{
  AthlonChips_s	*p_athlon;

  printf( "  Supported Chipsets:\n" );
  for ( p_athlon = AthlonChips; p_athlon->name != NULL; p_athlon++ ) {
    printf( "\t%s\n", p_athlon->name );
  }
}

static void	usage( const char *package )
{
  printf( " Usage: %s option\n", package );
  printf( "  supported options\n" );
  printf( "    on    enable Athlon Powersaving mode\n" );
  printf( "    off   disable Athlon Powersaving mode\n" );
  printf( "    stat  check status Athlon Powersaving mode\n" );
  printf( "    list  display support chipsets list\n\n" );
}

int	setAthlonPowersave( int m, struct device *first_dev )
{
  unsigned int	regval_cur, regval_new, res;
  struct device	*d = NULL;
  const char	*msg;
  AthlonChips_s	*p_athlon = NULL;
  ChipReg_s	*p_reg;
  EachReg_s	*p_eachreg;

  /* search supported chipset */
  for ( d = first_dev; d != NULL; d = d->next ) {
    for ( p_athlon = AthlonChips; p_athlon->name != NULL ; p_athlon++ ) {
      if ( (d->dev->vendor_id == p_athlon->vendorid) &&
	   (d->dev->device_id == p_athlon->deviceid) ) {
	fprintf( stderr, "%s (%04x %04x) found\n", p_athlon->name,
		 d->dev->vendor_id, d->dev->device_id );
	break;
      }
    }
    if ( p_athlon->name != NULL ) {
      break;
    }
  }

  if ( p_athlon == NULL || p_athlon->name == NULL ) {
    fprintf( stderr, "no supported Chipset found. nothing changed.\n" );
    return 1;
  }

  /* search chipset config */
  for ( p_reg = ChipReg; p_reg->regs->desc != NULL; p_reg++ ) {
    if ( p_athlon->chipreg == p_reg->regid ) {
      break;
    }
  }
  if ( p_reg->regs->desc == NULL ) {
    fprintf( stderr,
	     "internal error."
	     " supported Chipset found,"
	     " but Chipset Configuration not found.\n");
    return 1;
  }

  /* read current value of register */
  for ( p_eachreg = p_reg->regs; p_eachreg->desc != NULL; p_eachreg++ ) {
    switch (p_reg->tp) {
    case BYTE:
      regval_cur = pci_read_byte( d->dev, p_eachreg->addr );
      break;
    case SHORT:
      regval_cur = pci_read_word( d->dev, p_eachreg->addr );
      break;
    case LONG:
    default:
      regval_cur = pci_read_long( d->dev, p_eachreg->addr );
      break;
    }

    switch (m) {
    case ATHLONPM_STAT:
      /* check status */
      fprintf( stderr, "'%s' bit is %s.\n", p_eachreg->desc,
	       ( (regval_cur & p_eachreg->bit) == p_eachreg->bit ) ?
	       "enabled" : "disabled" );
      break;
    case ATHLONPM_ON:
    case ATHLONPM_OFF:
      /* enabling/disabling bits */
      regval_new = regval_cur | p_eachreg->bit;
      msg = ( m == ATHLONPM_ON ) ? "enabl" : "disabl";
      fprintf( stderr, "%sing '%s' bit ... ", msg, p_eachreg->desc );

      if ( ((regval_cur == regval_new) && (m == ATHLONPM_ON)) ||
	   ((regval_cur != regval_new) && (m == ATHLONPM_OFF)) ) {
	fprintf( stderr, "already %sed.\n", msg );
	break;
      }

      if ( m == ATHLONPM_OFF ) {
	regval_new = regval_cur ^ p_eachreg->bit;
      }

#ifndef ATHLON_PM_READONLY
      /* write new value into register */
      switch (p_reg->tp) {
      case BYTE:
	res = pci_write_byte( d->dev, p_eachreg->addr, (u8) regval_new );
	fprintf( stderr, " done\n" );
	fprintf( stderr, " Address 0x%02X : 0x%02X -> 0x%02X\n", 
		 p_eachreg->addr, regval_cur, regval_new );
	break;
      case SHORT:
	res = pci_write_word( d->dev, p_eachreg->addr, (u16) regval_new );
	fprintf( stderr, " done\n" );
	fprintf( stderr, " Address 0x%02X : 0x%04X -> 0x%04X\n", 
		 p_eachreg->addr, regval_cur, regval_new );
	break;
      case LONG:
      default:
	res = pci_write_long( d->dev, p_eachreg->addr, (u32) regval_new );
	fprintf( stderr, " done\n" );
	fprintf( stderr, " Address 0x%02X : 0x%08X -> 0x%08X\n", 
		 p_eachreg->addr, regval_cur, regval_new );
	break;
      }

#endif
      break;
    }
  }
  return 0;
}

int	main ( int argc, char *argv[] )
{
  int	mode = ATHLONPM_USAGE;

  /* read optarg */
  if ( argc == 2 ) {
    if ( strcmp( argv[1], "on") == 0 ) {
      mode = ATHLONPM_ON;
    } else if ( strcmp( argv[1], "off" ) == 0 ) {
      mode = ATHLONPM_OFF;
    } else if ( strcmp( argv[1], "stat" ) == 0 ) {
      mode = ATHLONPM_STAT;
    } else if ( strcmp( argv[1], "list" ) == 0 ) {
      mode = ATHLONPM_LIST;
    }
  }

  print_title();
  switch (mode) {
  case ATHLONPM_USAGE:
    usage( argv[0] );
    exit (1);
  case ATHLONPM_LIST:
    list_chipsets();
    exit (1);
  }

  if ( iopl( 3 ) < 0 ) {
    fprintf( stderr, "%s : must run as root. exit\n", argv[0] );
    exit (1);
  }

  /* init pciutil, and scan devices */
  pacc = pci_alloc();
  pci_filter_init( pacc, &filter );
  pci_init( pacc );

  setAthlonPowersave( mode, scan_devices() );

  /* clean up, exit */
  pci_cleanup( pacc );

  return (0);
}
/* End Of File	**************************************************************/
